#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<stdarg.h>
#include<time.h>
#include<ctype.h>
#include <GL/glut.h>
#include "texture.h"
#include "laby.h"

int width, height, depth;
int SCside[2], SCA[2], SCB[2], SCX[2], SCY[2], SCZ[2];
static char *field, SCitem[2];
int *walls_x, *walls_y, *walls_z, walls_SC[2];

static struct gen{
  int x, y, z;
  char dir;
} *genfield;
static int genlength[2], genmax;

void warning(char *s, ...)
{
  va_list vl;

  fprintf(stderr, "warning: ");
  va_start(vl, s);
  vfprintf(stderr, s, vl);
  va_end(vl);
  fprintf(stderr, "\n");
}

void error(char *s, ...)
{
  va_list vl;

  fprintf(stderr, "error: ");
  va_start(vl, s);
  vfprintf(stderr, s, vl);
  va_end(vl);
  fprintf(stderr, "\n");
  exit(1);
}

static char *field_coor(int x, int y, int z)
{
  int i;
  if(x<0 || y<0 || z<0 || x>=width || y>=height || z>=depth){
    for(i=0; i<2; i++) if(x==SCX[i] && y==SCY[i] && z==SCZ[i])
			 return &SCitem[i];
    return NULL;
  }
  return &field[x+y*width+z*width*height];
}

char backdir(char dir)
{ return dir/2*2+(1-dir%2);}

int *wall_coor(int x, int y, int z, char dir)
{
  int i;
  int *walls, a,b,c, maxa,maxb,maxc;

  if(dir < 0 || dir > 5) return NULL;
  if(x<0 || y<0 || z<0 || x>=width || y>=height || z>=depth){
    for(i=0; i<2; i++)
      if(x==SCX[i] && y==SCY[i] && z==SCZ[i] &&
	 dir == backdir(SCside[i]))
	return &walls_SC[i];
    return NULL;
  }
  switch(dir/2){
  case 0:
    walls = walls_x; a=y; b=z; c=x; maxa=height; maxb=depth; maxc=width;
    break;
  case 1:
    walls = walls_y; a=x; b=z; c=y; maxa=width; maxb=depth; maxc=height;
    break;
  case 2:
    walls = walls_z; a=x; b=y; c=z; maxa=width; maxb=height; maxc=depth;
  }

  maxc--;
  if(dir%2) c--;
  if(c<0 || c >= maxc){
    for(i=0; i<2; i++)
      if(a == SCA[i] && b == SCB[i] && dir == SCside[i])
	return &walls_SC[i];
    return NULL;
  }
  return &walls[a+b*maxa+c*maxa*maxb];
}

void movecoor(int *x, int *y, int *z, char dir)
{
  int *c;

  c = NULL;
  switch(dir/2){
  case 0: c=x; break;  /* x: RIGHT, LEFT */
  case 1: c=y; break;  /* y: DOWN, UP */
  case 2: c=z; break;  /* z: FORWARD, BACKWARD */
  }
  if(!c){
    warning("unknown direction %d", dir);
    return;
  }
  if(dir%2) (*c)--;
  else (*c)++;
}

void printdir(char dir)
{
  switch(dir){
  case RIGHT:  printf("RIGHT\n"); break;
  case LEFT:   printf("LEFT\n"); break;
  case DOWN:    printf("DOWN\n"); break;
  case UP:  printf("UP\n"); break;
  case FORWARD: printf("FORWARD\n"); break;
  case BACKWARD:  printf("BACKWARD\n"); break;
  default: warning("unknown direction");
  }
}

static void init()
{
  int i, length;

  length = width*height*depth;
  field = (char *)malloc(length*sizeof(char));
  if(!field) error("memory allocation failed");
  for(i=0; i<length; i++) field[i] = 0;

  length = (width-1)*height*depth;
  walls_x = (int *)malloc(length*sizeof(int));
  if(!walls_x) error("memory allocation failed");

  length = width*(height-1)*depth;
  walls_y = (int *)malloc(length*sizeof(int));
  if(!walls_y) error("memory allocation failed");

  length = width*height*(depth-1);
  walls_z = (int *)malloc(length*sizeof(int));
  if(!walls_z) error("memory allocation failed");

  genmax = (width-1)*height*depth+width*(height-1)*depth
    +width*height*(depth-1)+2;
  genfield = (struct gen *)malloc(genmax*sizeof(struct gen));
  if(!genfield) error("memory allocation failed");

  for(i=0; i<2; i++){
    if(SCside[i] < 2){
      SCY[i]=SCA[i];
      SCZ[i]=SCB[i];
      if(SCside[i]==1) SCX[i] = -1;
      else SCX[i] = width;
    }
    else if(SCside[i] < 4){
      SCX[i]=SCA[i];
      SCZ[i]=SCB[i];
      if(SCside[i]==3) SCY[i] = -1;
      else SCY[i] = height;
    }
    else{
      SCX[i]=SCA[i];
      SCY[i]=SCB[i];
      if(SCside[i]==5) SCZ[i] = -1;
      else SCZ[i] = depth;
    }

    walls_SC[i] = i*(genmax-1);
    genfield[walls_SC[i]].x = SCX[i];
    genfield[walls_SC[i]].y = SCY[i];
    genfield[walls_SC[i]].z = SCZ[i];
    genfield[walls_SC[i]].dir = backdir(SCside[i]);
    SCitem[i] = 2*i-1;
    genlength[i] = 1;
  }
}

/*
static void printgen()
{
  int i;
  for(i=0; i<genlength; i++){
    printf("%d:  %d, %d, %d,  ", i,genfield[i].x,genfield[i].y,genfield[i].z);
    switch(genfield[i].dir){
    case RIGHT: printf("RIGHT\n");break;
    case LEFT: printf("LEFT\n");break;
    case DOWN: printf("DOWN\n");break;
    case UP: printf("UP\n");break;
    case FORWARD: printf("FORWARD\n");break;
    case BACKWARD: printf("BACKWARD\n");break;
    }
  }
}
*/

static void addgen(int x, int y, int z, char dir, int type)
{
  int i;

  if(genlength[0] + genlength[1] >= genmax){
    error("Bludi capacity exceeded!");
  }
  i = genlength[type];
  if(type) i = genmax-1-i;
  *wall_coor(x, y, z, dir) = i;
  genfield[i].x = x;
  genfield[i].y = y;
  genfield[i].z = z;
  genfield[i].dir = dir;
  genlength[type] ++;
}

static void removegen(int i, char zbyde)
{
  int type;
  int j;

  if(i < genlength[0]) type = 0;
  else type = 1;

  genlength[type]--;
  j = genlength[type];
  if(type) j = genmax-1-j;

  *wall_coor(genfield[j].x, genfield[j].y,
	    genfield[j].z, genfield[j].dir) = i;
  *wall_coor(genfield[i].x, genfield[i].y, genfield[i].z, 
	    genfield[i].dir) = zbyde;
  genfield[i] = genfield[j];
}

static int *freqfield;
static int blindways_num;
static int lengthlaby;
static int turns_min;

static void breakwall(int x, int y, int z, char dir)
{
  int i,length,lastdir;
  int x2, y2, z2, dir2;
  char *pol;

  *wall_coor(x, y, z, dir) = 0;

  if(*field_coor(x,y,z) > 0){
    movecoor(&x, &y, &z, dir);
    dir = backdir(dir);
  }

  if(*field_coor(x,y,z) < -6) blindways_num++;
  x2=x; y2=y; z2=z;
  movecoor(&x2,&y2,&z2,dir);
  if(*field_coor(x2,y2,z2) > 6) blindways_num++;  

  length = width*height*depth;
  for(i=0; i<length; i++){
    if(field[i] < 0) field[i] *= -1;
    if(field[i] >= 7) field[i] -= 7;
    else field[i]--;
    if(field[i] >= 6 || field[i] < 0) warning("strange number %d on item %d", i,field[i]);
  }

  turns_min = 0;
  lengthlaby = 1;

  x2=x; y2=y; z2=z;
  lastdir = backdir(dir);
  while(x2 != SCX[0] || y2 != SCY[0] || z2 != SCZ[0]){
    pol = field_coor(x2,y2,z2);
    if(!pol){
      warning("I didn't find the way to the start");
      return;
    }
    dir2 = *pol;
    *pol = 6+backdir(lastdir)+dir2*6;

    if(lastdir != dir2) turns_min++;
    lengthlaby++;

    movecoor(&x2, &y2, &z2, dir2);
    lastdir = dir2;
 }

  x2=x; y2=y; z2=z;
  movecoor(&x2, &y2, &z2, dir);
  lastdir = dir;
  while(x2 != SCX[1] || y2 != SCY[1] || z2 != SCZ[1]){
    pol = field_coor(x2,y2,z2);
    if(!pol){
      warning("I didn't find the way to the finish");
      return;
    }
    dir2 = *pol;
    *pol = 6+dir2+backdir(lastdir)*6;

    if(lastdir != dir2) turns_min++;
    lengthlaby++;

    movecoor(&x2, &y2, &z2, dir2);
    lastdir = dir2;
  }
}

static void breakSC()
{
  int x, y, z, i, j;
  char old, new;

  for(i=0; i<2; i++)
    if(walls_SC[i]){
      breakwall(SCX[i], SCY[i], SCZ[i], backdir(SCside[i]));
      return;
    }

  i=0;
  j=1;
  for(;;){
    for(x=0; x<width; x++)
      for(y=0; y<height; y++)
	for(z=0; z<depth; z++){
	  old = new;
	  new = *field_coor(x, y, z);
	  if(z > 0 && old*new < 0){
	    i += j;
	    if(!i){
	      breakwall(x, y, z, BACKWARD);
	      return;
	    }
	  }
	}
    for(x=0; x<width; x++)
      for(z=0; z<depth; z++)
	for(y=0; y<height; y++){
	  old = new;
	  new = *field_coor(x, y, z);
	  if(y > 0 && old*new < 0){
	    i += j;
	    if(!i){
	      breakwall(x, y, z, UP);
	      return;
	    }
	  }
	}
    for(y=0; y<height; y++)
      for(z=0; z<depth; z++)
	for(x=0; x<width; x++){
	  old = new;
	  new = *field_coor(x, y, z);
	  if(x > 0 && old*new < 0){
	    i += j;
	    if(!i){
	      breakwall(x, y, z, LEFT);
	      return;
	    }
	  }
	}

    if(i == 0){
      warning("the bounds not found");
      return;
    }
    if(j == -1){
      warning("the end not found");
      return;
    }
    i = rand()%i+1;
    j = -1;
  }
}

static void generate()
{
  int i, x, y, z, dir, x2, y2, z2, type;
  char *item1, *item2;
  int plusnum = 0;

  blindways_num = 0;

  while(genlength[0] || genlength[1]){
    if(genlength[0] && genlength[1]) type = rand()%2;
    else if(genlength[0]) type = 0;
    else type = 1;

    i = rand()%genlength[type];
    if(type) i = genmax-1-i;
    x = genfield[i].x;
    y = genfield[i].y;
    z = genfield[i].z;
    dir = genfield[i].dir;
    removegen(i, 0);

    item2 = field_coor(x, y, z);
    if(!item2) error("item2 == NULL");

    movecoor(&x, &y, &z, dir);
    dir = backdir(dir);

    item1 = field_coor(x, y, z);
    if(!item1) error("item1 == NULL");
    if(*item2 > 0){
      plusnum++;
      *item1 = dir+1;
      if(*item2 <= 6) *item2 += 6;
      else blindways_num++;
    }
    else{
      *item1 = -dir-1;
      if(*item2 >= -6) *item2 -= 6;
      else blindways_num++;
    }

    for(i=0; i<6; i++){
      if(i == dir) continue;
      x2 = x; y2 = y; z2 = z;
      movecoor(&x2, &y2, &z2, i);
      item2 = field_coor(x2, y2, z2);
      if(item2){
	if(*item2 == 0) addgen(x, y, z, i, type);
	else removegen(*wall_coor(x, y, z, i), 1);
      }
    }
  }
}

void createlaby()
{
  int i, x,y,z;

  if(width <=0 || height <= 0 || depth <= 0)
    error("bad size of the labyrinth: %dx%dx%d", width, height, depth);
  for(i=0; i<2; i++){
    if(SCside[i] < 0 || SCside[i] > 5)
      error("bad number of the side of the start or of the finish, have to be 0,1,2,3,4,5");
    if(SCside[i] < 2){x=0, y=SCA[i], z=SCB[i];}
    else if(SCside[i] < 4){x=SCA[i], y=0, z=SCB[i];}
    else {x=SCA[i], y=SCB[i], z=0;}
    if(x<0 || y<0 || z<0 || x>=width || y>=height || z>=depth)
      error("bad coordinates of the start or the finish");
  }
  if(SCside[0] == SCside[1] && SCA[0] == SCA[1] && SCB[0] == SCB[1])
    error("start = finish");


  init();
  generate();
  breakSC();

  free(genfield);
}

void printlaby()
{
  int x, y, z;
  char c;

  for(z = depth-1; z>=0; z--){
    for(y=-1; y<height; y++){
      if(y>=0){
	for(x=-1; x<width; x++){
	  if(x>=0){
	    c=' ';
	    if(!iswall(x,y,z, FORWARD)) c='O';
	    if(!iswall(x,y,z, BACKWARD)){
	      if(c==' ') c='.';
	      else c = '*';
	    }
	    printf("%c", c);
	  }
	  if(iswall(x,y,z, RIGHT)) printf("|");
	  else printf(" ");
	}
	printf("\n");
      }
      for(x=-1; x<width; x++){
	if(x>=0){
	  if(iswall(x,y,z, DOWN)) printf("-");
	  else printf(" ");
	}
	c = ' ';
	if(iswall(x,y,z, RIGHT) || iswall(x,y+1,z, RIGHT)) c = '|';
	if(iswall(x,y,z, DOWN) || iswall(x+1,y,z, DOWN)){
	  if(c == ' ') c = '-';
	  else c = '+';
	}
	printf("%c", c);
      }
      printf("\n");
    }
    if(z>0) printf("\n\n\n\n");
  }
}

void printhelp(char *name)
{
  printf("\n");
  printf("Usage: %s options\n\n", name);
  printf("  -h -? -help --help       this text\n");
  printf("\n");
  printf("  -s --size <A>    size is set to AxAxA (default A=4)\n");
  printf("            <A>x<B>  size is set to AxBx1\n");
  printf("            <A>x<B>x<C>  size is set to AxBxC\n");
  printf("\n");
  printf("  -S --seed <init-string for the random generator>\n");
  printf("        if the same size and this string is used then\n");
  printf("        the same labyrinth is generated\n");
  printf("\n");
  printf("  -l --level <number>\n");
  printf("     level->size:  1->1x1x1, 2->2x1x1 3->2x2x1, 4->2x2x2,\n");
  printf("                   5->3x2x2, 10->4x4x4, 20->8x7x7, ...\n");
  printf("\n");
  printf("  -d --demo    the program runs through labyrinth automatically\n");
  printf("\n");
  printf("  -f --fullscreen  full screen mode\n");
  printf("                   (key f while game is running)\n");
  printf("\n");
  printf("  -i --invert  interchange arrows up and down\n");
  printf("                   (key i while game is running)\n");
  printf("\n");
  printf("Keystrokes:\n");
  printf("  rotation by arrows, moving forward by space-key, exit game by Esc\n");
  printf("Goal:\n");
  printf("  go throught labyrinth from blue square to green one or vice versa.\n");
  printf("\n");
}

void paramerror()
{
  error("bad options, try -h");
}

void setlaby(int argc, char **argv, char *fstart)
{
  int i, j;
  unsigned int seed;

  width = height = depth = 4;

  srand(time(NULL));

  for(j=1; j < argc; j++){
    if(!strcmp("-h", argv[j]) || !strcmp("-?", argv[j])  ||
       !strcmp("-help", argv[j]) || !strcmp("--help", argv[j])){
      printhelp(argv[0]);
      exit(0);
    }
    else if(!strcmp("-s", argv[j]) || !strcmp("--size", argv[j])){
      j++;
      if(j == argc) paramerror();
      width = height = depth = 0;
      if(!isdigit(argv[j][0])) paramerror();
      for(i=0; isdigit(argv[j][i]); i++) width = width*10 + argv[j][i]-'0';
      if(!argv[j][i]){
	depth = height = width;
	continue;
      }
      if(argv[j][i] != 'x') paramerror();
      if(!isdigit(argv[j][++i])) paramerror();
      for(; isdigit(argv[j][i]); i++) height = height*10 + argv[j][i]-'0';
      if(!argv[j][i]){
	depth = 1;
	continue;
      }
      if(argv[j][i] != 'x') paramerror();
      if(!isdigit(argv[j][++i])) paramerror();
      for(; isdigit(argv[j][i]); i++) depth = depth*10 + argv[j][i]-'0';
      if(argv[j][i]) paramerror();
    }
    else if(!strcmp("-l", argv[j]) || !strcmp("--level", argv[j])){
      j++;
      if(j == argc) paramerror();
      width = 0;
      if(!isdigit(argv[j][0])) paramerror();
      for(i=0; isdigit(argv[j][i]); i++) width = width*10 + argv[j][i]-'0';
      if(argv[j][i]) paramerror();

      depth = (width+2)/3;
      height = (width+3)/3;
      width = (width+4)/3;
    }
    else if(!strcmp("-S", argv[j]) || !strcmp("--seed", argv[j])){
      j++;
      if(j == argc) paramerror();
      seed=i=0;
      for(i=0; argv[j][i]; i++) seed = (seed*16+(argv[j][i]))%(3999);
      srand(seed);
    }
    else if(!strcmp("-d", argv[j]) || !strcmp("--demo", argv[j]))
      demo = 2;
    else if(!strcmp("-f", argv[j]) || !strcmp("--fullscreen", argv[j]))
      *fstart = 1;
    else if(!strcmp("-i", argv[j]) || !strcmp("--invert", argv[j]))
      invert = 1;
    else paramerror();
  }

  printf ("dim: %dx%dx%d\n", width, height, depth);

  SCside[0] = SCside[1] = UP;
  SCA[0] = SCA[1] = 0;
  SCB[0] = SCB[1] = (depth-1);
  SCA[1]++;

  SCside[1] = DOWN;
  SCA[1] = width-1;
  SCB[1] = 0;
}

static int state;
static int numtry, maxfreq;
static int demodir;
static int kstartuX, kstartuY, kstartuZ;

void analyzelaby()
{
  int length, i, x,y,z;

  length = width*height*depth;

  freqfield = (int *)malloc(length*sizeof(int));
  if(!freqfield) error("memory allocation failed");
  for(i = 0; i<length; i++)
    freqfield[i] = 0;

  maxfreq = 0;
  state = 0; numtry = 0;

  kstartuX = SCX[0]+1; kstartuY = SCY[0]+1; kstartuZ = SCZ[0]+1;
  demodir = 0;

  x = SCX[1]+1; y = SCY[1]+1; z = SCZ[1]+1;
  if(abs(x)+abs(y)+abs(z) < abs(kstartuX)+abs(kstartuY)+abs(kstartuZ)){
    kstartuX = x;
    kstartuY = y;
    kstartuZ = z;
    demodir = 1;
  }
}

static int moves_num, turns_num, blindways_visited, lastdir;

char analyzetah(int x, int y, int z, char dir)
{
  int i, j;
  int x2, y2, z2;
  char winner;

  if(iswall(x,y,z,dir)) return 0;

  winner = 0;
  moves_num++;
  if(dir != lastdir){
    turns_num++;
    lastdir = dir;
  }

  i = x+y*width+z*width*height;
  x2 = x; y2 = y; z2 = z;
  movecoor(&x2,&y2,&z2,dir);

  if(x2<0 || y2<0 || z2<0 || x2>=width || y2>=height || z2>=depth){
    if(state < 0) return 1;
    if(x2 == SCX[state] && y2 == SCY[state] && z2 == SCZ[state]) winner = 1;
    state = -1;
    i = -1;
  }
  else{
    if(state < 0){
      if(x == SCX[0] && y == SCY[0] && z == SCZ[0]) state = 1;
      else state = 0;
      moves_num = 1; turns_num = 0; numtry++;
    }
  }

  if(state >= 0){
    j = x2+y2*width+z2*width*height;
    freqfield[j]++;
    if(freqfield[j] > maxfreq) maxfreq = freqfield[j];

    if(field[j] >= 0 && field[j] < 6){
      field[j] -= 6;
      if(i < 0){
	warning("blind way on the start");
	return 1;
      }
      if(field[i] >= -6 && field[i] < 0) field[i] -= 6;
      else blindways_visited++;
    }
  }
  if(winner){
    printf("YOU WON\n");
    printf("You have tried it %d times\n", numtry);
    if(blindways_num == 0) i = 0;
    else i = 100*blindways_visited/blindways_num;
    printf("You have visited %d from %d blind passages, it is %d%%\n",
	   blindways_visited, blindways_num, i);
    printf("The most frequent place was visited %d times\n", maxfreq);
    printf("You have used %d moves and %d turns.\n",
	   moves_num, turns_num);
    if(moves_num == 0) i = 100;
    else i = 100*lengthlaby/moves_num;
    if(turns_num == 0) j = 100;
    else j = 100*turns_min/turns_num;
    printf("Ideally, to solve this labyrinth is sufficient %d (%d%%) moves and %d (%d%%) turns\n",
	   lengthlaby, i, turns_min, j);
    return 2;
  }
  return 1;
}

void demotest()
{
  int x,y,z, dir;

  dir=x=y=z=-1;

  for(;;){
    dir = demomove(x,y,z);
    if(dir == -1) break;
    printdir(dir);
    movecoor(&x,&y,&z,dir);
  }
}

char demomove(int x, int y, int z)
{
  if(kstartuY > 0){
    kstartuY--;
    return DOWN;
  }
  if(kstartuX+x < kstartuZ+z)
    if(kstartuZ > 0){
      kstartuZ--;
      return FORWARD;
    }
  if(kstartuX > 0){
    kstartuX--;
    return RIGHT;
  }
  if(kstartuZ > 0){
    kstartuZ--;
    return FORWARD;
  }
  if(x == SCX[1-demodir] && y == SCY[1-demodir] && z == SCZ[1-demodir]) return -1;
  if(x == SCX[demodir] && y == SCY[demodir] && z == SCZ[demodir])
    return backdir(SCside[demodir]);

  if(demodir == 0) return (*field_coor(x,y,z))%6;
  else return (*field_coor(x,y,z))/6-1;
}

/*
int main(int argc, char ** argv)
{
  setlaby();
  createlaby();
  printlaby();

  return 0;  
}
*/
